==> Couchbase-Java-Client-3.10.0.zip: com/couchbase/client/core/msg/analytics/AnalyticsRequest.java <==
/*
 * Copyright (c) 2019 Couchbase, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.couchbase.client.core.msg.analytics;

import com.couchbase.client.core.CoreContext;
import com.couchbase.client.core.cnc.CbTracing;
import com.couchbase.client.core.cnc.RequestSpan;
import com.couchbase.client.core.cnc.TracingIdentifiers;
import com.couchbase.client.core.deps.io.netty.buffer.ByteBuf;
import com.couchbase.client.core.deps.io.netty.buffer.Unpooled;
import com.couchbase.client.core.deps.io.netty.handler.codec.http.DefaultFullHttpRequest;
import com.couchbase.client.core.deps.io.netty.handler.codec.http.FullHttpRequest;
import com.couchbase.client.core.deps.io.netty.handler.codec.http.HttpHeaderNames;
import com.couchbase.client.core.deps.io.netty.handler.codec.http.HttpHeaderValues;
import com.couchbase.client.core.deps.io.netty.handler.codec.http.HttpMethod;
import com.couchbase.client.core.deps.io.netty.handler.codec.http.HttpVersion;
import com.couchbase.client.core.env.Authenticator;
import com.couchbase.client.core.msg.BaseRequest;
import com.couchbase.client.core.msg.HttpRequest;
import com.couchbase.client.core.msg.ResponseStatus;
import com.couchbase.client.core.retry.RetryStrategy;
import com.couchbase.client.core.service.ServiceType;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.time.Duration;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import static com.couchbase.client.core.logging.RedactableArgument.redactMeta;
import static com.couchbase.client.core.logging.RedactableArgument.redactUser;
import static com.couchbase.client.core.util.CbCollections.listOf;

public class AnalyticsRequest
  extends BaseRequest<AnalyticsResponse>
  implements HttpRequest<AnalyticsChunkHeader, AnalyticsChunkRow, AnalyticsChunkTrailer, AnalyticsResponse> {

  public static final int NO_PRIORITY = 0;

  private static final List<String> httpPathsByVersion = listOf(
    "/analytics/service", // v0
    "/api/v1/request"
  );

  private static final HttpMethod httpMethod = HttpMethod.POST;

  private final byte[] query;
  private final int priority;
  private final boolean idempotent;
  private final String contextId;
  private final String statement;
  private final String bucket;
  private final String scope;
  private final String httpPath;

  private final Authenticator authenticator;

  /**
   * Automatically translate error codes into specific SDK3 exceptions,
   * or throw generic {@link com.couchbase.client.core.error.CoreErrorCodeAndMessageException}?
   */
  private final boolean translateExceptions;

  public AnalyticsRequest(final Duration timeout, final CoreContext ctx, final RetryStrategy retryStrategy,
                          final Authenticator authenticator, final byte[] query, final int priority,
                          final boolean idempotent, final String contextId, final String statement,
                          final RequestSpan span, final String bucket, final String scope) {
    this(timeout, ctx, retryStrategy, authenticator, query, priority, idempotent, contextId, statement, span, bucket,
      scope, true, 0);
  }

  public AnalyticsRequest(
    final Duration timeout,
    final CoreContext ctx,
    final RetryStrategy retryStrategy,
    final Authenticator authenticator,
    final byte[] query,
    final int priority,
    final boolean idempotent,
    final String contextId,
    final String statement,
    final RequestSpan span,
    final String bucket,
    final String scope,
    final boolean translateExceptions,
    final int apiVersion
  ) {
    super(timeout, ctx, retryStrategy, span);
    this.query = query;
    this.authenticator = authenticator;
    this.priority = priority;
    this.idempotent = idempotent;
    this.contextId = contextId;
    this.statement = statement;
    this.bucket = bucket;
    this.scope = scope;
    this.translateExceptions = translateExceptions;
    this.httpPath = httpPathsByVersion.get(apiVersion);

    if (span != null && !CbTracing.isInternalSpan(span)) {
      span.lowCardinalityAttribute(TracingIdentifiers.ATTR_SERVICE, TracingIdentifiers.SERVICE_ANALYTICS);
      span.attribute(TracingIdentifiers.ATTR_STATEMENT, statement);
      if (bucket != null) {
        span.attribute(TracingIdentifiers.ATTR_NAME, bucket);
      }
      if (scope != null) {
        span.attribute(TracingIdentifiers.ATTR_SCOPE, scope);
      }
    }
  }

  /**
   * Helper method to build the query context from bucket and scope.
   *
   * @param bucket the name of the bucket.
   * @param scope the name of the scope.
   * @return the build query context string.
   */
  public static String queryContext(final String bucket, final String scope) {
    return "default:`"  + bucket + "`.`" + scope +"`" ;
  }

  @Override
  public FullHttpRequest encode() {
    ByteBuf content = query == null || query.length == 0 ? Unpooled.EMPTY_BUFFER : Unpooled.wrappedBuffer(query);
    FullHttpRequest request = new DefaultFullHttpRequest(HttpVersion.HTTP_1_1, httpMethod,
      httpPath, content);
    request.headers()
      .set(HttpHeaderNames.CONTENT_TYPE, HttpHeaderValues.APPLICATION_JSON);
    request.headers()
      .set(HttpHeaderNames.CONTENT_LENGTH, content.readableBytes());
    request.headers()
      .set(HttpHeaderNames.USER_AGENT, context().environment().userAgent().formattedLong());

    if (priority != NO_PRIORITY) {
      request.headers().set("Analytics-Priority", priority);
    }

    authenticator.authHttpRequest(serviceType(), request);
    return request;
  }

  @Override
  public AnalyticsResponse decode(final ResponseStatus status, final AnalyticsChunkHeader header,
                                  final Flux<AnalyticsChunkRow> rows,
                                  final Mono<AnalyticsChunkTrailer> trailer) {
    return new AnalyticsResponse(status, header, rows, trailer);
  }

  @Override
  public ServiceType serviceType() {
    return ServiceType.ANALYTICS;
  }

  @Override
  public boolean idempotent() {
    return idempotent;
  }

  @Override
  public String operationId() {
    return contextId;
  }

  public String statement() {
    return statement;
  }

  public boolean translateExceptions() {
    return translateExceptions;
  }

  @Override
  public Map<String, Object> serviceContext() {
    Map<String, Object> ctx = new TreeMap<>();
    ctx.put("type", serviceType().ident());
    ctx.put("httpPath", httpPath);
    ctx.put("httpMethod", httpMethod.toString());
    ctx.put("operationId", redactMeta(operationId()));
    ctx.put("statement", redactUser(statement()));
    ctx.put("priority", priority);
    if (bucket != null) {
      ctx.put("bucket", redactMeta(bucket));
    }
    if (scope != null) {
      ctx.put("scope", redactMeta(scope));
    }
    return ctx;
  }

  public String bucket() {
    return bucket;
  }

  public String scope() {
    return scope;
  }

  @Override
  public String name() {
    return "analytics";
  }

}

==> Couchbase-Java-Client-3.11.1.zip: com/couchbase/client/core/msg/analytics/AnalyticsRequest.java <==
/*
 * Copyright (c) 2019 Couchbase, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.couchbase.client.core.msg.analytics;

import com.couchbase.client.core.CoreContext;
import com.couchbase.client.core.annotation.UsedBy;
import com.couchbase.client.core.annotation.UsedBy.Project;
import com.couchbase.client.core.cnc.CbTracing;
import com.couchbase.client.core.cnc.RequestSpan;
import com.couchbase.client.core.cnc.TracingIdentifiers;
import com.couchbase.client.core.cnc.tracing.TracingAttribute;
import com.couchbase.client.core.cnc.tracing.TracingDecorator;
import com.couchbase.client.core.deps.io.netty.buffer.ByteBuf;
import com.couchbase.client.core.deps.io.netty.buffer.Unpooled;
import com.couchbase.client.core.deps.io.netty.handler.codec.http.DefaultFullHttpRequest;
import com.couchbase.client.core.deps.io.netty.handler.codec.http.FullHttpRequest;
import com.couchbase.client.core.deps.io.netty.handler.codec.http.HttpHeaderNames;
import com.couchbase.client.core.deps.io.netty.handler.codec.http.HttpHeaderValues;
import com.couchbase.client.core.deps.io.netty.handler.codec.http.HttpMethod;
import com.couchbase.client.core.deps.io.netty.handler.codec.http.HttpVersion;
import com.couchbase.client.core.env.Authenticator;
import com.couchbase.client.core.msg.BaseRequest;
import com.couchbase.client.core.msg.HttpRequest;
import com.couchbase.client.core.msg.ResponseStatus;
import com.couchbase.client.core.retry.RetryStrategy;
import com.couchbase.client.core.service.ServiceType;
import org.jspecify.annotations.Nullable;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.time.Duration;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import static com.couchbase.client.core.logging.RedactableArgument.redactMeta;
import static com.couchbase.client.core.logging.RedactableArgument.redactUser;
import static com.couchbase.client.core.util.CbCollections.listOf;

public class AnalyticsRequest
  extends BaseRequest<AnalyticsResponse>
  implements HttpRequest<AnalyticsChunkHeader, AnalyticsChunkRow, AnalyticsChunkTrailer, AnalyticsResponse> {

  public static final int NO_PRIORITY = 0;

  private static final List<String> httpPathsByVersion = listOf(
    "/analytics/service", // v0
    "/api/v1/request"
  );

  private final byte[] query;
  private final int priority;
  private final boolean idempotent;
  private final String contextId;
  private final String statement;
  private final String bucket;
  private final String scope;
  private final String httpPath;
  private final HttpMethod httpMethod;

  private final Authenticator authenticator;

  /**
   * Automatically translate error codes into specific SDK3 exceptions,
   * or throw generic {@link com.couchbase.client.core.error.CoreErrorCodeAndMessageException}?
   */
  private final boolean translateExceptions;

  /**
   * "Canonical" constructor that just sets all the properties.
   */
  private AnalyticsRequest(
    final Duration timeout,
    final CoreContext ctx,
    final RetryStrategy retryStrategy,
    final Authenticator authenticator,
    final RequestSpan span,
    final byte[] query,
    final int priority,
    final boolean idempotent,
    final String contextId,
    final String statement,
    final String bucket,
    final String scope,
    final String httpPath,
    final HttpMethod httpMethod,
    final boolean translateExceptions
  ) {
    super(timeout, ctx, retryStrategy, span);

    this.query = query;
    this.priority = priority;
    this.idempotent = idempotent;
    this.contextId = contextId;
    this.statement = statement;
    this.bucket = bucket;
    this.scope = scope;
    this.httpPath = httpPath;
    this.httpMethod = httpMethod;
    this.authenticator = authenticator;
    this.translateExceptions = translateExceptions;

    if (span != null && !CbTracing.isInternalSpan(span)) {
      TracingDecorator tip = ctx.coreResources().tracingDecorator();
      tip.provideLowCardinalityAttr(TracingAttribute.SERVICE, span, TracingIdentifiers.SERVICE_ANALYTICS);
      tip.provideAttr(TracingAttribute.STATEMENT, span, statement);
      if (bucket != null) {
        tip.provideLowCardinalityAttr(TracingAttribute.BUCKET_NAME, span, bucket);
      }
      if (scope != null) {
        tip.provideAttr(TracingAttribute.SCOPE_NAME, span, scope);
      }
    }
  }

  /**
   * Constructor for JDBC driver so it can specify the HTTP path and method.
   */
  @UsedBy(Project.JDBC_DRIVER)
  public AnalyticsRequest(
    final Duration timeout,
    final CoreContext ctx,
    final RetryStrategy retryStrategy,
    final Authenticator authenticator,
    final byte @Nullable [] query,
    final int priority,
    final boolean idempotent,
    final String contextId,
    final String statement,
    final RequestSpan span,
    final String bucket,
    final String scope,
    final String httpPath,
    final HttpMethod httpMethod
  ) {
    this(
      timeout,
      ctx,
      retryStrategy,
      authenticator,
      span,
      query,
      priority,
      idempotent,
      contextId,
      statement,
      bucket,
      scope,
      httpPath,
      httpMethod,
      true
    );
  }

  /**
   * Constructor for Operational SDK (legacy Analytics).
   */
  public AnalyticsRequest(
    final Duration timeout,
    final CoreContext ctx,
    final RetryStrategy retryStrategy,
    final Authenticator authenticator,
    final byte[] query,
    final int priority,
    final boolean idempotent,
    final String contextId,
    final String statement,
    final RequestSpan span,
    final String bucket,
    final String scope
  ) {
    this(
      timeout,
      ctx,
      retryStrategy,
      authenticator,
      query,
      priority,
      idempotent,
      contextId,
      statement,
      span,
      bucket,
      scope,
      true,
      0
    );
  }

  /**
   * Constructor for Columnar SDK so it can specify API version and disable exception translation.
   */
  public AnalyticsRequest(
    final Duration timeout,
    final CoreContext ctx,
    final RetryStrategy retryStrategy,
    final Authenticator authenticator,
    final byte[] query,
    final int priority,
    final boolean idempotent,
    final String contextId,
    final String statement,
    final RequestSpan span,
    final String bucket,
    final String scope,
    final boolean translateExceptions,
    final int apiVersion
  ) {
    this(
      timeout,
      ctx,
      retryStrategy,
      authenticator,
      span,
      query,
      priority,
      idempotent,
      contextId,
      statement,
      bucket,
      scope,
      httpPathsByVersion.get(apiVersion),
      HttpMethod.POST,
      translateExceptions
    );
  }

  /**
   * Helper method to build the query context from bucket and scope.
   *
   * @param bucket the name of the bucket.
   * @param scope the name of the scope.
   * @return the build query context string.
   */
  public static String queryContext(final String bucket, final String scope) {
    return "default:`"  + bucket + "`.`" + scope +"`" ;
  }

  @Override
  public FullHttpRequest encode() {
    ByteBuf content = query == null || query.length == 0 ? Unpooled.EMPTY_BUFFER : Unpooled.wrappedBuffer(query);
    FullHttpRequest request = new DefaultFullHttpRequest(HttpVersion.HTTP_1_1, httpMethod,
      httpPath, content);
    request.headers()
      .set(HttpHeaderNames.CONTENT_TYPE, HttpHeaderValues.APPLICATION_JSON);
    request.headers()
      .set(HttpHeaderNames.CONTENT_LENGTH, content.readableBytes());
    request.headers()
      .set(HttpHeaderNames.USER_AGENT, context().environment().userAgent().formattedLong());

    if (priority != NO_PRIORITY) {
      request.headers().set("Analytics-Priority", priority);
    }

    authenticator.authHttpRequest(serviceType(), request);
    return request;
  }

  @Override
  public AnalyticsResponse decode(final ResponseStatus status, final AnalyticsChunkHeader header,
                                  final Flux<AnalyticsChunkRow> rows,
                                  final Mono<AnalyticsChunkTrailer> trailer) {
    return new AnalyticsResponse(status, header, rows, trailer);
  }

  @Override
  public ServiceType serviceType() {
    return ServiceType.ANALYTICS;
  }

  @Override
  public boolean idempotent() {
    return idempotent;
  }

  @Override
  public String operationId() {
    return contextId;
  }

  public String statement() {
    return statement;
  }

  public boolean translateExceptions() {
    return translateExceptions;
  }

  @Override
  public Map<String, Object> serviceContext() {
    Map<String, Object> ctx = new TreeMap<>();
    ctx.put("type", serviceType().ident());
    ctx.put("httpPath", httpPath);
    ctx.put("httpMethod", httpMethod.toString());
    ctx.put("operationId", redactMeta(operationId()));
    ctx.put("statement", redactUser(statement()));
    ctx.put("priority", priority);
    if (bucket != null) {
      ctx.put("bucket", redactMeta(bucket));
    }
    if (scope != null) {
      ctx.put("scope", redactMeta(scope));
    }
    return ctx;
  }

  public String bucket() {
    return bucket;
  }

  public String scope() {
    return scope;
  }

  @Override
  public String name() {
    return "analytics";
  }

}
